/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file command.c Client authentication and server commands. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <signal.h>
#include <sys/types.h>
#include <osl.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "server.lsg.h"
#include "para.h"
#include "error.h"
#include "crypt.h"
#include "sideband.h"
#include "command.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "list.h"
#include "send.h"
#include "sched.h"
#include "vss.h"
#include "net.h"
#include "daemon.h"
#include "fd.h"
#include "ipc.h"
#include "server_cmd.lsg.h"
#include "user_list.h"
#include "signal.h"
#include "version.h"

#define SERVER_CMD_AUX_INFO(_arg) _arg,
static const unsigned server_command_perms[] = {LSG_SERVER_CMD_AUX_INFOS};
#undef SERVER_CMD_AUX_INFO
#define SERVER_CMD_AUX_INFO(_arg) #_arg,
static const char * const server_command_perms_txt[] = {LSG_SERVER_CMD_AUX_INFOS};
#undef SERVER_CMD_AUX_INFO

/** Commands including options must be shorter than this. */
#define MAX_COMMAND_LEN 32768

extern int mmd_mutex;
extern struct misc_meta_data *mmd;
extern struct sender senders[];
int send_afs_status(struct command_context *cc, int parser_friendly);

static void dummy(__a_unused int s)
{
}

/*
 * Compute human readable vss status text.
 *
 * We can't call vss_playing() and friends here because those functions read
 * the flags from the primary mmd structure, so calling them from command
 * handler context would require to take the mmd lock. At the time the function
 * is called we already took a copy of the mmd structure and want to use the
 * flags value of the copy for computing the vss status text.
 */
static char *vss_status_tohuman(unsigned int flags)
{
	if (flags & VSS_PLAYING)
		return para_strdup("playing");
	if (flags & VSS_NEXT)
		return para_strdup("stopped");
	return para_strdup("paused");
}

/*
 * Never returns NULL.
 */
static char *vss_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & VSS_PLAYING)? 'P' : '_';
	msg[1] = (flags & VSS_NOMORE)? 'O' : '_';
	msg[2] = (flags & VSS_NEXT)? 'N' : '_';
	msg[3] = (flags & VSS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

static unsigned get_status(struct misc_meta_data *nmmd, bool parser_friendly,
		char **result)
{
	char *status, *flags; /* vss status info */
	/* nobody updates our version of "now" */
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval current_time;
	struct para_buffer b = {.flags = parser_friendly? PBF_SIZE_PREFIX : 0};

	/* report real status */
	status = vss_status_tohuman(nmmd->vss_status_flags);
	flags = vss_get_status_flags(nmmd->vss_status_flags);
	clock_get_realtime(&current_time);
	/*
	 * The calls to WRITE_STATUS_ITEM() below never fail because
	 * b->max_size is zero (unlimited), see \ref para_printf(). However,
	 * clang is not smart enough to prove this and complains nevertheless.
	 * Casting the return value to void silences clang.
	 */
	(void)WRITE_STATUS_ITEM(&b, SI_STATUS, "%s\n", status);
	(void)WRITE_STATUS_ITEM(&b, SI_STATUS_FLAGS, "%s\n", flags);
	(void)WRITE_STATUS_ITEM(&b, SI_OFFSET, "%li\n", offset);
	(void)WRITE_STATUS_ITEM(&b, SI_AFS_MODE, "%s\n", mmd->afs_mode_string);
	(void)WRITE_STATUS_ITEM(&b, SI_STREAM_START, "%lu.%lu\n",
		(long unsigned)nmmd->stream_start.tv_sec,
		(long unsigned)nmmd->stream_start.tv_usec);
	(void)WRITE_STATUS_ITEM(&b, SI_CURRENT_TIME, "%lu.%lu\n",
		(long unsigned)current_time.tv_sec,
		(long unsigned)current_time.tv_usec);
	free(flags);
	free(status);
	*result = b.buf;
	return b.offset;
}

/**
 * Send a sideband packet through a blocking file descriptor.
 *
 * \param scc fd and crypto keys.
 * \param buf The buffer to send.
 * \param numbytes The size of \a buf.
 * \param band The sideband designator of this packet.
 * \param dont_free If true, never deallocate \a buf.
 *
 * The nonblock flag must be disabled for the file descriptor given by \a scc.
 *
 * Stream cipher encryption is automatically activated if necessary via the
 * sideband transformation, depending on the value of \a band.
 *
 * \return Standard.
 *
 * \sa \ref send_sb_va().
 */
int send_sb(struct stream_cipher_context *scc, void *buf, size_t numbytes,
		int band, bool dont_free)
{
	int ret;
	struct sb_context *sbc;
	struct iovec iov[2];
	sb_transformation trafo = band < SBD_PROCEED? NULL : sc_trafo;
	struct sb_buffer sbb = SBB_INIT(band, buf, numbytes);

	sbc = sb_new_send(&sbb, dont_free, trafo, scc->send);
	do {
		ret = sb_get_send_buffers(sbc, iov);
		ret = xwritev(scc->fd, iov, ret);
		if (ret < 0)
			goto fail;
	} while (sb_sent(sbc, ret) == false);
	return 1;
fail:
	sb_free(sbc);
	return ret;
}

/**
 * Create a variable sized buffer and send it as a sideband packet.
 *
 * \param scc Passed to \ref send_sb.
 * \param band See \ref send_sb.
 * \param fmt The format string.
 *
 * \return The return value of the underlying call to \ref send_sb.
 */
__printf_3_4 int send_sb_va(struct stream_cipher_context *scc, int band,
		const char *fmt, ...)
{
	va_list ap;
	char *msg;
	int ret;

	va_start(ap, fmt);
	ret = xvasprintf(&msg, fmt, ap);
	va_end(ap);
	return send_sb(scc, msg, ret, band, false);
}

/**
 * Send an error message to a client.
 *
 * \param cc Client info.
 * \param err The (positive) error code.
 *
 * \return The return value of the underlying call to send_sb_va().
 */
int send_strerror(struct command_context *cc, int err)
{
	return send_sb_va(&cc->scc, SBD_ERROR_LOG, "%s\n", para_strerror(err));
}

/**
 * Send an error context to a client,
 *
 * \param cc Client info.
 * \param errctx The error context string.
 *
 * \return The return value of the underlying call to send_sb_va().
 *
 * This function frees the error context string after it was sent.
 */
int send_errctx(struct command_context *cc, char *errctx)
{
	int ret;

	if (!errctx)
		return 0;
	ret = send_sb_va(&cc->scc, SBD_ERROR_LOG, "%s\n", errctx);
	free(errctx);
	return ret;
}

static int check_sender_args(struct command_context *cc,
		struct lls_parse_result *lpr, struct sender_command_data *scd)
{
	int i, ret;
	const char * const subcmds[] = {SENDER_SUBCOMMANDS};
	const char *arg;
	char *errctx;
	unsigned num_inputs = lls_num_inputs(lpr);

	scd->sender_num = -1;
	ret = lls(lls_check_arg_count(lpr, 2, INT_MAX, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	arg = lls_input(0, lpr);
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, arg))
			break;
	if (!senders[i].name)
		return -E_COMMAND_SYNTAX;
	scd->sender_num = i;
	arg = lls_input(1, lpr);
	for (i = 0; i < NUM_SENDER_CMDS; i++)
		if (!strcmp(subcmds[i], arg))
			break;
	if (i == NUM_SENDER_CMDS)
		return -E_COMMAND_SYNTAX;
	scd->cmd_num = i;
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num])
		return -E_SENDER_CMD;
	switch (scd->cmd_num) {
	case SENDER_on:
	case SENDER_off:
		if (num_inputs != 2)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_deny:
	case SENDER_allow:
		if (num_inputs != 3 || parse_cidr(lls_input(2, lpr), scd->host,
				sizeof(scd->host), &scd->netmask) == NULL)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_add:
	case SENDER_delete:
		if (num_inputs != 3)
			return -E_COMMAND_SYNTAX;
		return parse_fec_url(lls_input(2, lpr), scd);
	default:
		return -E_COMMAND_SYNTAX;
	}
	return 1;
}

/**
 * Send a sideband packet through a blocking file descriptor.
 *
 * \param scc fd and crypto keys.
 * \param expected_band The expected band designator.
 * \param max_size Passed to \ref sb_new_recv().
 * \param result Body of the sideband packet is returned here.
 *
 * If \a expected_band is not \p SBD_ANY, the band designator of the received
 * sideband packet is compared to \a expected_band and a mismatch is considered
 * an error.
 *
 * \return Standard.
 */
int recv_sb(struct stream_cipher_context *scc,
		enum sb_designator expected_band,
		size_t max_size, struct iovec *result)
{
	int ret;
	struct sb_context *sbc;
	struct iovec iov;
	struct sb_buffer sbb;
	sb_transformation trafo;

	trafo = expected_band != SBD_ANY && expected_band < SBD_PROCEED?
		NULL : sc_trafo;
	sbc = sb_new_recv(max_size, trafo, scc->recv);
	for (;;) {
		sb_get_recv_buffer(sbc, &iov);
		ret = recv_bin_buffer(scc->fd, iov.iov_base, iov.iov_len);
		if (ret == 0)
			ret = -E_EOF;
		if (ret < 0)
			goto fail;
		ret = sb_received(sbc, ret, &sbb);
		if (ret < 0)
			goto fail;
		if (ret > 0)
			break;
	}
	ret = -E_BAD_BAND;
	if (expected_band != SBD_ANY && sbb.band != expected_band)
		goto fail;
	*result = sbb.iov;
	return 1;
fail:
	sb_free(sbc);
	return ret;
}

static int com_sender(struct command_context *cc, struct lls_parse_result *lpr)
{
	int i, ret = 0;
	char *msg = NULL;
	struct sender_command_data scd;

	if (lls_num_inputs(lpr) == 0) {
		for (i = 0; senders[i].name; i++) {
			char *tmp;
			ret = xasprintf(&tmp, "%s%s\n", msg? msg : "",
				senders[i].name);
			free(msg);
			msg = tmp;
		}
		return send_sb(&cc->scc, msg, ret, SBD_OUTPUT, false);
	}
	ret = check_sender_args(cc, lpr, &scd);
	if (ret < 0) {
		if (scd.sender_num < 0)
			return ret;
		if (strcmp(lls_input(1, lpr), "status") == 0)
			msg = senders[scd.sender_num].status();
		else
			msg = senders[scd.sender_num].help();
		return send_sb(&cc->scc, msg, strlen(msg), SBD_OUTPUT, false);
	}

	switch (scd.cmd_num) {
	case SENDER_add:
	case SENDER_delete:
		assert(senders[scd.sender_num].resolve_target);
		ret = senders[scd.sender_num].resolve_target(lls_input(2, lpr),
			&scd);
		if (ret < 0)
			return ret;
	}

	for (i = 0; i < 10; i++) {
		mutex_lock(mmd_mutex);
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			/* another sender command is active, retry in 100ms */
			struct timespec ts = {.tv_nsec = 100 * 1000 * 1000};
			mutex_unlock(mmd_mutex);
			nanosleep(&ts, NULL);
			continue;
		}
		mmd->sender_cmd_data = scd;
		mutex_unlock(mmd_mutex);
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}
EXPORT_SERVER_CMD_HANDLER(sender);

static int com_si(struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	char *msg, *ut;
	int ret;

	ut = daemon_get_uptime_str(now);
	mutex_lock(mmd_mutex);
	ret = xasprintf(&msg,
		"up: %s\nplayed: %u\n"
		"server_pid: %d\n"
		"afs_pid: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %s\n"
		"supported audio formats: %s\n",
		ut, mmd->num_played,
		(int)getppid(),
		(int)mmd->afs_pid,
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		ENUM_STRING_VAL(LOGLEVEL),
		AUDIO_FORMAT_HANDLERS
	);
	mutex_unlock(mmd_mutex);
	free(ut);
	return send_sb(&cc->scc, msg, ret, SBD_OUTPUT, false);
}
EXPORT_SERVER_CMD_HANDLER(si);

static int com_version(struct command_context *cc, struct lls_parse_result *lpr)
{
	char *msg;
	size_t len;

	if (SERVER_CMD_OPT_GIVEN(VERSION, VERBOSE, lpr))
		len = xasprintf(&msg, "%s", version_text("server"));
	else
		len = xasprintf(&msg, "%s\n", version_single_line("server"));
	return send_sb(&cc->scc, msg, len, SBD_OUTPUT, false);
}
EXPORT_SERVER_CMD_HANDLER(version);

/** These status items are cleared if no audio file is currently open. */
#define EMPTY_STATUS_ITEMS \
	ITEM(PATH) \
	ITEM(DIRECTORY) \
	ITEM(BASENAME) \
	ITEM(SCORE) \
	ITEM(ATTRIBUTES_BITMAP) \
	ITEM(ATTRIBUTES_TXT) \
	ITEM(HASH) \
	ITEM(IMAGE_ID) \
	ITEM(IMAGE_NAME) \
	ITEM(LYRICS_ID) \
	ITEM(LYRICS_NAME) \
	ITEM(BITRATE) \
	ITEM(FORMAT) \
	ITEM(FREQUENCY) \
	ITEM(CHANNELS) \
	ITEM(DURATION) \
	ITEM(SECONDS_TOTAL) \
	ITEM(NUM_PLAYED) \
	ITEM(LAST_PLAYED) \
	ITEM(TECHINFO) \
	ITEM(ARTIST) \
	ITEM(TITLE) \
	ITEM(YEAR) \
	ITEM(ALBUM) \
	ITEM(COMMENT) \
	ITEM(MTIME) \
	ITEM(FILE_SIZE) \
	ITEM(CHUNK_TIME) \
	ITEM(NUM_CHUNKS) \
	ITEM(AMPLIFICATION) \

/*
 * Create a set of audio-file related status items with empty values. These are
 * written to stat clients when no audio file is open.
 */
static unsigned empty_status_items(bool parser_friendly, char **result)
{
	char *esi;
	unsigned len;

	if (parser_friendly)
		len = xasprintf(&esi,
			#define ITEM(x) "0004 %02x:\n"
			EMPTY_STATUS_ITEMS
			#undef ITEM
			#define ITEM(x) , (unsigned) SI_ ## x
			EMPTY_STATUS_ITEMS
			#undef ITEM
		);
	else
		len = xasprintf(&esi,
			#define ITEM(x) "%s:\n"
			EMPTY_STATUS_ITEMS
			#undef ITEM
			#define ITEM(x) ,status_item_list[SI_ ## x]
			EMPTY_STATUS_ITEMS
			#undef ITEM
		);
	*result = esi;
	return len;
}
#undef EMPTY_STATUS_ITEMS

static int com_stat(struct command_context *cc, struct lls_parse_result *lpr)
{
	int ret;
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;
	bool parser_friendly = SERVER_CMD_OPT_GIVEN(STAT, PARSER_FRIENDLY,
		lpr) > 0;
	uint32_t num = SERVER_CMD_UINT32_VAL(STAT, NUM, lpr);

	para_sigaction(SIGUSR1, dummy);
	for (;;) {
		/*
		 * Copy the mmd structure to minimize the time we hold the mmd
		 * lock.
		 */
		mutex_lock(mmd_mutex);
		*nmmd = *mmd;
		mutex_unlock(mmd_mutex);
		ret = get_status(nmmd, parser_friendly, &s);
		ret = send_sb(&cc->scc, s, ret, SBD_OUTPUT, false);
		if (ret < 0)
			goto out;
		if (nmmd->vss_status_flags & VSS_NEXT) {
			char *esi;
			ret = empty_status_items(parser_friendly, &esi);
			ret = send_sb(&cc->scc, esi, ret, SBD_OUTPUT, false);
			if (ret < 0)
				goto out;
		} else
			send_afs_status(cc, parser_friendly);
		ret = 1;
		if (num > 0 && !--num)
			goto out;
		sleep(50);
		ret = -E_SERVER_CRASH;
		if (getppid() == 1)
			goto out;
	}
out:
	return ret;
}
EXPORT_SERVER_CMD_HANDLER(stat);

/* fixed-length, human readable permission string */
const char *server_cmd_perms_str(unsigned int perms)
{
	static char result[5];

	result[0] = perms & AFS_READ? 'a' : '-';
	result[1] = perms & AFS_WRITE? 'A' : '-';
	result[2] = perms & VSS_READ? 'v' : '-';
	result[3] = perms & VSS_WRITE? 'V' : '-';
	result[4] = '\0';
	return result;
}

static int send_list_of_commands(struct command_context *cc)
{
	int i;
	const struct lls_command *cmd;
	char *msg = para_strdup("");

	for (i = 1; (cmd = lls_cmd(i, server_cmd_suite)); i++) {
		const char *perms = server_cmd_perms_str(server_command_perms[i]);
		char *tmp = make_message("%s%s\t%s\t%s\n", msg,
			lls_command_name(cmd), perms, lls_purpose(cmd));
		free(msg);
		msg = tmp;
	}
	return send_sb(&cc->scc, msg, strlen(msg), SBD_OUTPUT, false);
}

static int com_help(struct command_context *cc, struct lls_parse_result *lpr)
{
	const char *perms;
	char *long_help, *buf, *errctx;
	int ret;
	const struct lls_command *cmd;

	ret = lls(lls_check_arg_count(lpr, 0, 1, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	if (lls_num_inputs(lpr) == 0)
		return send_list_of_commands(cc);
	/* argument given for help */
	ret = lls(lls_lookup_subcmd(lls_input(0, lpr), server_cmd_suite,
		&errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	cmd = lls_cmd(ret, server_cmd_suite);
	perms = server_command_perms_txt[ret];
	long_help = lls_long_help(cmd);
	assert(long_help);
	ret = xasprintf(&buf, "%spermissions: %s\n", long_help, perms);
	free(long_help);
	return send_sb(&cc->scc, buf, ret, SBD_OUTPUT, false);
}
EXPORT_SERVER_CMD_HANDLER(help);

static int com_hup(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	kill(getppid(), SIGHUP);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(hup);

static int com_term(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	kill(getppid(), SIGTERM);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(term);

static int com_play(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(play);

static int com_stop(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(stop);

static int com_pause(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	mutex_lock(mmd_mutex);
	if (!vss_paused() && !vss_stopped()) {
		mmd->events++;
		mmd->new_vss_status_flags &= ~VSS_PLAYING;
		mmd->new_vss_status_flags &= ~VSS_NEXT;
	}
	mutex_unlock(mmd_mutex);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(pause);

static int com_next(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	mutex_lock(mmd_mutex);
	mmd->events++;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(next);

static int com_nomore(__a_unused struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	mutex_lock(mmd_mutex);
	if (vss_playing() || vss_paused())
		mmd->new_vss_status_flags |= VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}
EXPORT_SERVER_CMD_HANDLER(nomore);

static int com_ff(__a_unused struct command_context *cc,
		struct lls_parse_result *lpr)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c, *errctx;

	ret = lls(lls_check_arg_count(lpr, 1, 1, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	if (!(ret = sscanf(lls_input(0, lpr), "%u%c", &i, &c)))
		return -E_COMMAND_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total || !mmd->afd.afhi.seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->afd.afhi.chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->afd.afhi.seconds_total;
	else
		promille += 1000 * i / mmd->afd.afhi.seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille > 1000) {
		mmd->new_vss_status_flags |= VSS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->afd.afhi.chunks_total * promille) / 1000;
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}
EXPORT_SERVER_CMD_HANDLER(ff);

static int com_jmp(__a_unused struct command_context *cc,
		struct lls_parse_result *lpr)
{
	long unsigned int i;
	int ret;
	char *errctx;

	ret = lls(lls_check_arg_count(lpr, 1, 1, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	if (sscanf(lls_input(0, lpr), "%lu", &i) <= 0)
		return -ERRNO_TO_PARA_ERROR(EINVAL);
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->afd.afhi.chunks_total * i + 50) / 100;
	PARA_INFO_LOG("sent: %lu, offset before jmp: %li\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}
EXPORT_SERVER_CMD_HANDLER(jmp);

static int com_tasks(struct command_context *cc,
		__a_unused struct lls_parse_result *lpr)
{
	char *tl = server_get_tasks();
	assert(tl);
	return send_sb(&cc->scc, tl, strlen(tl), SBD_OUTPUT, false);
}
EXPORT_SERVER_CMD_HANDLER(tasks);

static void reset_signals(void)
{
	para_sigaction(SIGCHLD, SIG_IGN);
	para_sigaction(SIGINT, SIG_DFL);
	para_sigaction(SIGTERM, SIG_DFL);
	para_sigaction(SIGHUP, SIG_DFL);
}

struct connection_features {
	int dummy; /* none at the moment */
};

static int parse_auth_request(char *buf, int len, struct user **u,
		struct connection_features *cf)
{
	int ret;
	char *p, *username, **features = NULL;
	size_t auth_rq_len = strlen(AUTH_REQUEST_MSG);

	*u = NULL;
	memset(cf, 0, sizeof(*cf));
	if (len < auth_rq_len + 2)
		return -E_AUTH_REQUEST;
	if (strncmp(buf, AUTH_REQUEST_MSG, auth_rq_len) != 0)
		return -E_AUTH_REQUEST;
	username = buf + auth_rq_len;
	p = strchr(username, ' ');
	if (p) {
		int i;
		if (p == username)
			return -E_AUTH_REQUEST;
		*p = '\0';
		p++;
		create_argv(p, ",", &features);
		for (i = 0; features[i]; i++) {
			if (strcmp(features[i], "sideband") == 0)
				continue;
			if (strcmp(features[i], "aes_ctr128") == 0)
				continue;
			else {
				ret = -E_BAD_FEATURE;
				goto out;
			}
		}
	}
	PARA_DEBUG_LOG("received auth request for user %s\n", username);
	*u = lookup_user(username);
	ret = 1;
out:
	free_argv(features);
	return ret;
}

#define HANDSHAKE_BUFSIZE 4096

static int run_command(struct command_context *cc, struct iovec *iov,
		const char *peername)
{
	int ret, i, argc;
	char *p, *end, **argv;
	const struct lls_command *lcmd = NULL;
	unsigned perms;
	struct lls_parse_result *lpr;
	char *errctx;

	if (iov->iov_base == NULL || iov->iov_len == 0)
		return -ERRNO_TO_PARA_ERROR(EINVAL);
	p = iov->iov_base;
	p[iov->iov_len - 1] = '\0'; /* just to be sure */

	ret = lls(lls_lookup_subcmd(p, server_cmd_suite, &errctx));
	if (ret < 0) {
		send_errctx(cc, errctx);
		return ret;
	}
	perms = server_command_perms[ret];
	if ((perms & cc->u->perms) != perms)
		return -E_PERM;
	lcmd = lls_cmd(ret, server_cmd_suite);
	end = iov->iov_base + iov->iov_len;
	for (i = 0; p < end; i++)
		p += strlen(p) + 1;
	argc = i;
	argv = para_malloc((argc + 1) * sizeof(char *));
	for (i = 0, p = iov->iov_base; p < end; i++) {
		argv[i] = para_strdup(p);
		p += strlen(p) + 1;
	}
	argv[argc] = NULL;
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", lls_command_name(lcmd),
		cc->u->name, peername);
	ret = lls(lls_parse(argc, argv, lcmd, &lpr, &errctx));
	if (ret >= 0) {
		const struct server_cmd_user_data *ud = lls_user_data(lcmd);
		ret = ud->handler(cc, lpr);
		lls_free_parse_result(lpr, lcmd);
	} else
		send_errctx(cc, errctx);
	free_argv(argv);
	mutex_lock(mmd_mutex);
	mmd->num_commands++;
	if (ret >= 0 && (perms & AFS_WRITE))
		mmd->events++;
	mutex_unlock(mmd_mutex);
	return ret;
}

/**
 * Perform user authentication and execute a command.
 *
 * \param fd The file descriptor to send output to.
 * \param peername Identifies the connecting peer.
 *
 * Whenever para_server accepts an incoming tcp connection on the port it
 * listens on, it forks and the resulting child calls this function.
 *
 * An RSA-based challenge/response is used to authenticate the peer. It that
 * authentication succeeds, a random session key is generated and sent back to
 * the peer, encrypted with its RSA public key. From this point on, all
 * transfers are crypted with this session key.
 *
 * Next it is checked if the peer supplied a valid server command or a command
 * for the audio file selector. If yes, and if the user has sufficient
 * permissions to execute that command, the function calls the corresponding
 * command handler which does argument checking and further processing.
 *
 * In order to cope with a DOS attacks, a timeout is set up which terminates
 * the function if the connection was not authenticated when the timeout
 * expires.
 *
 * \sa alarm(2), \ref crypt.c, \ref crypt.h.
 */
__noreturn void handle_connect(int fd, const char *peername)
{
	int ret;
	unsigned char rand_buf[CHALLENGE_SIZE + 2 * SESSION_KEY_LEN];
	unsigned char challenge_hash[HASH_SIZE];
	char *command = NULL, *buf = para_malloc(HANDSHAKE_BUFSIZE) /* must be on the heap */;
	size_t numbytes;
	struct command_context cc_struct = {.u = NULL}, *cc = &cc_struct;
	struct iovec iov;
	struct connection_features cf;

	cc->scc.fd = fd;
	reset_signals();
	/* we need a blocking fd here as recv() might return EAGAIN otherwise. */
	ret = mark_fd_blocking(fd);
	if (ret < 0)
		goto net_err;
	/* send Welcome message */
	ret = write_va_buffer(fd, "This is para_server, version "
		PACKAGE_VERSION ".\n"
		"Features: sideband,aes_ctr128\n"
	);
	if (ret < 0)
		goto net_err;
	/* recv auth request line */
	ret = recv_buffer(fd, buf, HANDSHAKE_BUFSIZE);
	if (ret < 0)
		goto net_err;
	ret = parse_auth_request(buf, ret, &cc->u, &cf);
	if (ret < 0)
		goto net_err;
	if (cc->u) {
		get_random_bytes_or_die(rand_buf, sizeof(rand_buf));
		ret = pub_encrypt(cc->u->pubkey, rand_buf, sizeof(rand_buf),
			(unsigned char *)buf);
		if (ret < 0)
			goto net_err;
		numbytes = ret;
	} else {
		/*
		 * We don't want to reveal our user names, so we send a
		 * challenge to the client even if the user does not exist, and
		 * fail the authentication later.
		 */
		numbytes = 256;
		get_random_bytes_or_die((unsigned char *)buf, numbytes);
	}
	PARA_DEBUG_LOG("sending %d byte challenge + session key (%zu bytes)\n",
		CHALLENGE_SIZE, numbytes);
	ret = send_sb(&cc->scc, buf, numbytes, SBD_CHALLENGE, false);
	buf = NULL;
	if (ret < 0)
		goto net_err;
	ret = recv_sb(&cc->scc, SBD_CHALLENGE_RESPONSE,
		HANDSHAKE_BUFSIZE, &iov);
	if (ret < 0)
		goto net_err;
	buf = iov.iov_base;
	numbytes = iov.iov_len;
	PARA_DEBUG_LOG("received %zu bytes challenge response\n", numbytes);
	ret = -E_BAD_USER;
	if (!cc->u)
		goto net_err;
	/*
	 * The correct response is the hash of the first CHALLENGE_SIZE bytes
	 * of the random data.
	 */
	ret = -E_BAD_AUTH;
	if (numbytes != HASH_SIZE)
		goto net_err;
	hash_function((char *)rand_buf, CHALLENGE_SIZE, challenge_hash);
	if (memcmp(challenge_hash, buf, HASH_SIZE))
		goto net_err;
	/* auth successful */
	alarm(0);
	PARA_INFO_LOG("good auth for %s\n", cc->u->name);
	/* init stream cipher keys with the second part of the random buffer */
	cc->scc.recv = sc_new(rand_buf + CHALLENGE_SIZE, SESSION_KEY_LEN);
	cc->scc.send = sc_new(rand_buf + CHALLENGE_SIZE + SESSION_KEY_LEN,
		SESSION_KEY_LEN);
	ret = send_sb(&cc->scc, NULL, 0, SBD_PROCEED, false);
	if (ret < 0)
		goto net_err;
	ret = recv_sb(&cc->scc, SBD_COMMAND, MAX_COMMAND_LEN, &iov);
	if (ret < 0)
		goto net_err;
	ret = run_command(cc, &iov, peername);
	free(iov.iov_base);
	if (ret < 0)
		goto err_out;
	if (ret >= 0)
		goto out;
err_out:
	if (send_strerror(cc, -ret) >= 0)
		send_sb(&cc->scc, NULL, 0, SBD_EXIT__FAILURE, true);
net_err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
out:
	free(buf);
	free(command);
	mutex_lock(mmd_mutex);
	mmd->active_connections--;
	mutex_unlock(mmd_mutex);
	if (ret >= 0) {
		ret = send_sb(&cc->scc, NULL, 0, SBD_EXIT__SUCCESS, true);
		if (ret < 0)
			PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
	}
	sc_free(cc->scc.recv);
	sc_free(cc->scc.send);
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
