/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter.h Filter-related structures and exported symbols from filter_common.c. */

/**
 * Describes one running instance of a filter.
*/
struct filter_node {
	/** The number in the array of available filters. */
	unsigned filter_num;
	/**
	 * Each filter may store any filter-specific information about the particular
	 * instance of the filter here.
	 */
	void *private_data;
	/** The list of registered callbacks. */
	struct list_head callbacks;
	/** A pointer to the configuration of this instance. */
	void *conf;
	/** The parsed command line, merged with options given in the config file. */
	struct lls_parse_result *lpr;
	/** The buffer tree node. */
	struct btr_node *btrn;
	/** The task corresponding to this filter node. */
	struct task *task;
	/** The minimal input queue size, see \ref btr_node_status(). */
	size_t min_iqs;
};

/**
 * The structure associated with a paraslash filter.
 *
 * Paraslash filters are "modules" which transform an audio stream. struct
 * filter contains methods which are implemented by each filter.
 *
 * Note: As several instances of the same filter may be running at the same
 * time, all these filter functions must be reentrant; no static non-constant
 * variables may be used.
 *
 * \sa \ref filter_node.
 */
struct filter {
	/**
	 * Open one instance of this filter.
	 *
	 * This should allocate the output buffer of the given filter node and
	 * do any other filter-specific preparations like initializing the
	 * private_data member of \a fn suitably. The open function is
	 * optional, If it is provided, it is assumed to succeed.
	 */
	void (*open)(struct filter_node *fn);
	/**
	 * Close one instance of this filter.
	 *
	 * Free all resources associated with \a fn that were previously
	 * allocated by the open() function. It's OK to set this to NULL if the
	 * filter does not need to perform any cleanup operation.
	 */
	void (*close)(struct filter_node *fn);
	/**
	 * Prepare the filter according to command line options.
	 *
	 * In addition to the syntactic checks which are automatically performed
	 * by the lopsub functions, some filters like to also check the command
	 * line arguments semantically. Moreover, since applications may open
	 * the filter many times with the same options, filters need a method
	 * which allows them to precompute once those parts of the setup which
	 * depend only on the command line options.
	 *
	 * If this function pointer is not NULL, the function is called once at
	 * startup. The returned pointer value is made available to the ->open
	 * method via the ->conf pointer of struct filter_node.
	 *
	 * Filters are supposed to abort if the setup fails. If the function
	 * returns, it is assumed to have succeeded.
	 */
	void *(*setup)(const struct lls_parse_result *lpr);
	/**
	 * Deallocate precomputed resources.
	 *
	 * This should free whatever ->setup() has allocated.
	 */
	void (*teardown)(const struct lls_parse_result *lpr, void *conf);
	/**
	 * Set scheduler timeout and add file descriptors to fd sets.
	 *
	 * This function controls the timeout value for the next call to
	 * select(2). It may decrease the current timeout but shall never
	 * increase it. The second purpose of this function is to add file
	 * descriptors to the two fd sets of the sched structure. The
	 * descriptors in these sets will be watched by the subsequent
	 * select(2) call.
	 */
	void (*pre_select)(struct sched *s, void *context);
	/**
	 * Convert (filter) the given data.
	 *
	 * Pointer to the converting function of the filter. On errors, the
	 * post_select function is supposed to return a negative error code.
	 */
	int (*post_select)(struct sched *s, void *context);
	/**
	 * Answer a buffer tree query.
	 *
	 * This optional function pointer is used for inter node communications
	 * of the buffer tree nodes. See \ref btr_command_handler for details.
	 */
	btr_command_handler execute;
};

void print_filter_helps(bool detailed);
void print_filter_list(void);
int filter_setup(const char *fa, void **conf, struct lls_parse_result **lprp);
#define FILTER_CMD(_num) (lls_cmd(_num, filter_cmd_suite))
#define FILTER_CMD_OPT(_cmd, _opt) (lls_opt( \
	LSG_FILTER_CMD_ ## _cmd ## _OPT_ ## _opt, \
	FILTER_CMD(LSG_FILTER_CMD_CMD_ ## _cmd)))
#define FILTER_CMD_OPT_RESULT(_cmd, _opt, _lpr) \
	(lls_opt_result(LSG_FILTER_CMD_ ## _cmd ## _OPT_ ## _opt, _lpr))
#define FILTER_CMD_OPT_GIVEN(_cmd, _opt, _lpr) \
	(lls_opt_given(FILTER_CMD_OPT_RESULT(_cmd, _opt, _lpr)))
#define FILTER_CMD_OPT_UINT32_VAL(_cmd, _opt, _lpr) \
	(lls_uint32_val(0, FILTER_CMD_OPT_RESULT(_cmd, _opt, _lpr)))
#define FILTER_CMD_OPT_STRING_VAL(_cmd, _opt, _lpr) \
	(lls_string_val(0, FILTER_CMD_OPT_RESULT(_cmd, _opt, _lpr)))

void generic_filter_pre_select(struct sched *s, void *context);
int decoder_execute(const char *cmd, unsigned sample_rate, unsigned channels,
		char **result);

static inline void write_int16_host_endian(char *buf, int val)
{
#ifdef WORDS_BIGENDIAN
	*buf = val >> 8;
	*(buf + 1) = val & 0xff;
#else
	*buf = val & 0xff;
	*(buf + 1) = val >> 8;
#endif
}

const struct filter *filter_get(int filter_num);
const char *filter_name(int filter_num);
